(****************************************************************************)
(* Level 0/1 Prover                                                         *)
(* Copyright (C) 2004 Alwen Tiu                                             *)
(*                                                                          *)
(* This program is free software; you can redistribute it and/or modify     *)
(* it under the terms of the GNU General Public License as published by     *)
(* the Free Software Foundation; either version 2 of the License, or        *)
(* (at your option) any later version.                                      *)
(*                                                                          *)
(* This program is distributed in the hope that it will be useful,          *)
(* but WITHOUT ANY WARRANTY; without even the implied warranty of           *)
(* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            *)
(* GNU General Public License for more details.                             *)
(*                                                                          *)
(* You should have received a copy of the GNU General Public License        *)
(* along with this code; if not, write to the Free Software Foundation,     *)
(* Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA             *)
(****************************************************************************)

(* Interface to the lambda term parser *)

signature PARSERINT =
sig
    structure Term : TERM
    structure Cls : CLAUSE

    datatype command = exit | load of string | query of Term.term list
    datatype preamble = incl of string
    datatype select = ParseCommand of command | ParseGoal of Term.term
              | ParseClause of Cls.clause | ParsePreamble of preamble
              
	val reset_symTbl : unit -> unit
	val lookup : (string * Term.tag) -> Term.term      
end

functor PIntFunct(structure Term : TERM 
		structure Cls : CLAUSE
		sharing Cls.Term = Term) : PARSERINT = 
struct
     structure Term = Term
     structure Cls = Cls

     datatype command = exit | load of string | query of Term.term list
     datatype preamble = incl of string
     datatype select = ParseCommand of command | ParseGoal of Term.term
              | ParseClause of Cls.clause | ParsePreamble of preamble


	val tbl : Term.term list ref = ref nil
	

	fun reset_symTbl() = (tbl := nil)
	  
	fun create_newid(s, flag) =
    	let val (c::_) = explode(s)
   		in
        	if (Char.isUpper c) then 
           	(ref (Term.var(s, 0, flag)))
        	else (ref (Term.const(s,0,flag)))
    	end

    
	fun lookup(s, flag) =
	    let fun search(s, nil) = NONE
	          | search(s, (t as ref (Term.const(u,_,_)))::ts) =
	               if (s = u) then (SOME t) 
		       else search(s, ts)
	          | search(s, (t as ref (Term.var(u,_,_)))::ts) =
	               if (s = u) then (SOME t)
	               else search(s, ts)         
	        val c = search(s, !tbl)
	    in
		case c of
	          SOME t => t
	        | NONE => 
	            let val newc = create_newid(s, flag)
		    in
	               (Term.checkAndUpdateName s; tbl := newc::(!tbl); newc)
		    end
	    end              
end
