(****************************************************************************)
(* Level 0/1 Prover                                                         *)
(* Copyright (C) 2004 Alwen Tiu                                             *)
(*                                                                          *)
(* This program is free software; you can redistribute it and/or modify     *)
(* it under the terms of the GNU General Public License as published by     *)
(* the Free Software Foundation; either version 2 of the License, or        *)
(* (at your option) any later version.                                      *)
(*                                                                          *)
(* This program is distributed in the hope that it will be useful,          *)
(* but WITHOUT ANY WARRANTY; without even the implied warranty of           *)
(* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            *)
(* GNU General Public License for more details.                             *)
(*                                                                          *)
(* You should have received a copy of the GNU General Public License        *)
(* along with this code; if not, write to the Free Software Foundation,     *)
(* Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA             *)
(****************************************************************************)

(* 
pprint.sml: pretty printer.
*)

signature PPRINT =
sig
	structure Term : TERM
	structure UStream : USTREAM
	structure Sub : SUBST
	
	val print_term : Term.term -> unit
	val print_abs_term  : int -> Term.term -> unit
	val print_abs_list : int -> Term.term list -> unit
	val print_list : Term.term list -> unit
	val present_substs  : (Sub.subst UStream.ustream) * Term.term list -> unit
	val print_substs : Sub.subst -> unit
	val writeFile : string * string * Term.term list -> unit
end

functor PPrintFunc(
			structure Term : TERM
			structure Norm : NORM
			structure Sub : SUBST
			structure UStream : USTREAM
			sharing Norm.Term = Term and Sub.Term = Term) : PPRINT =
struct
	structure Term = Term
	structure Norm = Norm
	structure Sub = Sub
	structure UStream = UStream
	
	open Term
	open Norm
	open Sub
	open UStream

	fun flatten_infix c nil = nil
	  | flatten_infix c (x::ts) =
	  	  (
	  		case x of (t as ref (app(ref (const(d,_,_)), rs))) =>
	  					(
	  						if (c = d) then 
	  							List.concat[(flatten_infix c rs),
	  										(flatten_infix c ts)]
	  						else (t::(flatten_infix c ts))
	  					)
				| t => (t::(flatten_infix c ts))
		  )
		  
	fun flatten (t as ref (var(S,i,f))) = t
	  | flatten (t as ref (const(S,i,f))) = t
	  | flatten (t as ref (dB(_))) = t
	  | flatten (ref (app(ref (app(h, ts1)), ts2))) =
			flatten (ref (app(h, List.concat[ts1,ts2]))) 
	  | flatten (ref (app(h as ref (const("&",_,_)), ts))) = 
	  		if (ts = nil) then h
	  		else ref (app(h, (flatten_infix "&" ts)))
	  | flatten (ref (app(h as ref (const(";",_,_)), ts))) = 
	  		if (ts = nil) then h
	  		else ref (app(h, (flatten_infix ";" ts)))	  		
	  | flatten (ref (app(h, ts))) =
			ref (app(flatten h, map flatten ts))
	  | flatten (ref (lam(n,t))) = ref (lam(n, flatten t))
	  | flatten t = flatten (hnorm t)
	  		  		
	fun fprint_abs Out n pre (ref (var(S,i,_))) = TextIO.output(Out, S)
  		| fprint_abs Out n pre (ref (const(S,i,f))) = TextIO.output(Out, S)
  		| fprint_abs Out n pre (ref (dB i)) = 
                     (TextIO.output(Out, pre); 
                     TextIO.output(Out, (Int.toString (n-i+1))))
		| fprint_abs Out n pre (ref (app(ref (const("&",_,_)), ts))) = 
		     ( TextIO.output(Out, "("); 
		     fprint_infix Out n pre "&" ts; 
		     TextIO.output(Out, ")") )
		| fprint_abs Out n pre (ref (app(ref (const(";",_,_)), ts))) = 
		     ( TextIO.output(Out, "("); 
		       fprint_infix Out n pre ";" ts; 
		       TextIO.output(Out, ")") )		     
		| fprint_abs Out n pre (ref (app(ref (const("=>",_,_)), ts))) = 
		     ( TextIO.output(Out, "("); 
		     fprint_infix Out n pre "=>" ts; 
		     TextIO.output(Out, ")") )			     	     
  		| fprint_abs Out n pre (ref (app(h, ts))) =
   	   		(TextIO.output(Out, "("); 
   	   		 fprint_abs Out n pre h ; 
   	   		 TextIO.output(Out, " "); 
   	   		 fprint_abs_ls Out n pre ts; 
   	   		 TextIO.output(Out, ")")
   	   		)
		| fprint_abs Out n pre (ref (lam(0,t))) = fprint_abs Out n pre t
  		| fprint_abs Out n pre (ref (lam(i, t))) =
    		  (
    		  TextIO.output(Out, pre); 
    		  TextIO.output(Out, (Int.toString (n+1))); 
    		  TextIO.output(Out, "\\"); 
              fprint_abs Out (n+1) pre (ref (lam(i-1,t)))
              )
	and fprint_abs_ls Out n pre nil = ()
  		| fprint_abs_ls Out n pre (t::rs) = (
  			fprint_abs Out n pre t; 
  			(if (rs = nil) then () else TextIO.output(Out, " ")); 
  			fprint_abs_ls Out n pre rs
  		   )
	and fprint_infix Out n pre s nil = ()
	  | fprint_infix Out n pre s (t::nil) = fprint_abs Out n pre t	       
	  | fprint_infix Out n pre s (t::ts) =
	  		(fprint_abs Out n pre t; 
	  		 TextIO.output(Out, " "^s^" "); 
	  		 fprint_infix Out n pre s ts
	  		 )  		   
  		   

	fun print_abs_term n t = fprint_abs (TextIO.stdOut) n (getAbsName()) (flatten t)
	fun print_abs_list n ts = fprint_abs_ls (TextIO.stdOut) n (getAbsName()) (map flatten ts)
	
    fun print_term t = print_abs_term 0 t
	fun print_list ts = print_abs_list 0 ts

	fun fprint_abs_term Out n t = fprint_abs Out n (getAbsName()) (flatten t)
	fun fprint_abs_list Out n ts = fprint_abs_ls Out n (getAbsName()) (map flatten ts)
	
	fun fprint_term Out t = fprint_abs_term Out 0 t
	fun fprint_list Out ts = fprint_abs_list Out 0 ts
	
   fun flush_white_space() =
         let val  next = TextIO.lookahead TextIO.stdIn 
         in
             if next = (SOME #"\n") orelse next = (SOME #" ")
             then (TextIO.input1 TextIO.stdIn; flush_white_space())
             else ()
         end
         
  	fun reply () =
       (flush_white_space();
        let val key = TextIO.input1 TextIO.stdIn  
        in
          if key = (SOME #"y") orelse key = (SOME #"Y")
                    orelse key = (SOME #".") then true else false
        end)
        
   fun show_subst(subst, vars) =
       (map (fn v => (print_term(v);
                      print " = ";
                      print_term(norm (apply_sub subst v));
                      print "\n"))
             vars)
               		   
   fun present_substs(substs, freevars) =
       if snull(substs) then print "No.\n" else
          (print "Yes\n";
           show_subst(head substs, freevars);
          print "\nFind another? [y/n] ";
          if reply() then present_substs(tail substs, freevars)
                  else print "No.\n")  		   

   fun print_substs(Subst L) = ((map (fn (v,t) => 
   					(print_term(v); 
   					print " = "; 
					print_term(norm t);
					print "\n")
				      )
				   L) ; ())
					 
	fun fprint_def Out t =
		let val (_,V,_) = get_prefixes()
		
			fun makeApp n t = 
				ref (app(t, [ref (var(V ^ (Int.toString n), 0, none))]))
				
			fun pdef k (ref (app(ref (const("def",_,_)), [H, B]))) =
					(
						fprint_term Out H; 
						TextIO.output(Out, " := "); 
						fprint_term Out B; 
						TextIO.output(Out, ".\n")
					)
			  | pdef k (s as ref (lam(n,r))) =
			  		if (n <= 0) then pdef k r
			  		else pdef (k+1) (norm (makeApp k s))
			  | pdef k r = (fprint_term Out t; TextIO.output(Out, ".\n"))
		in
			pdef 0 (norm t)
		end
		
	fun writeFile(path, fname, ts) =
		let val fpath = if (path = "" orelse (String.isSuffix "/" path) orelse 
							(String.isSuffix "\\" path)) then path
						else path^"/"
			val file = fpath^fname
			val Out = (SOME (TextIO.openOut(file)))
					handle IO.Io(_) => 
						(	print "Failed in opening file "; 
							print file; 
							print "\n"; NONE
						)
		in
			case Out of NONE => ()
			| (SOME OutStream) => 
				(map (fprint_def OutStream) ts; TextIO.closeOut(OutStream))
				handle IO.Io(_) =>
				(
					print "Write failed.\n";
					TextIO.closeOut(OutStream)
				)
		end                		
		
end

